# -*- coding: utf-8 -*-
import mock
from django.test import TestCase
from django.conf import settings
from django.contrib.auth import get_user_model
from social.apps.django_app.default.models import UserSocialAuth
from oauth_tokens.factories import AccessTokenFactory, UserCredentialsFactory
from oauth_tokens.models import AccessTokenGettingError
from vkontakte_api.api import VkontakteApi

from .api import override_api_context
from .exceptions import CallsLimitError


TOKEN = 'b492c0a63455412b67c579422119da1bf73ce07e3bf28f18fa8446c2441844eee57232ca15b7229122dd2'


def raise_error(*args, **kwargs):
    raise AccessTokenGettingError


class SocialApiUnitTest(TestCase):

    def test_override_api_context(self):
        with self.settings(SOCIAL_API_CALL_CONTEXT={}):
            self.assertEqual(settings.SOCIAL_API_CALL_CONTEXT, {})
            with override_api_context('vkontakte', user=1):
                self.assertEqual(settings.SOCIAL_API_CALL_CONTEXT, {'vkontakte': {'user': 1}})
                with override_api_context('vkontakte', token='abc'):
                    self.assertEqual(settings.SOCIAL_API_CALL_CONTEXT, {'vkontakte': {'user': 1, 'token': 'abc'}})

    def test_social_auth_user_argument(self):
        user = get_user_model().objects.create(username='user')
        for i in range(0, 10):
            AccessTokenFactory(provider='vkontakte')
        UserSocialAuth.objects.create(user=get_user_model().objects.create(), uid=2, provider='vk-oauth2',
                                      extra_data='{"access_token": "qwewekrjhshe", "expires": null, "id": null}')
        UserSocialAuth.objects.create(user=user, uid=1, provider='vk-oauth2',
                                      extra_data='{"access_token": "%s", "expires": null, "id": null}' % TOKEN)

        with override_api_context('vkontakte', social_auth_user=user):
            api = VkontakteApi()
            (self.assertEqual(api.get_token(), TOKEN) for i in range(0, 10))

    def test_oauth_tokens_tag_argument(self):
        user_cr = UserCredentialsFactory()
        user_cr.tags.add('tag')
        AccessTokenFactory(provider='vkontakte', access_token=TOKEN, user_credentials=user_cr)
        for i in range(0, 10):
            AccessTokenFactory(provider='vkontakte')
        UserSocialAuth.objects.create(user=get_user_model().objects.create(), uid=1, provider='vk-oauth2',
                                      extra_data='{"access_token": "qwewekrjhshe", "expires": null, "id": null}')

        with override_api_context('vkontakte', oauth_tokens_tag='tag'):
            api = VkontakteApi()
            (self.assertEqual(api.get_token(), TOKEN) for i in range(0, 10))

    def test_get_tokens(self):
        for i in range(0, 5):
            token = AccessTokenFactory(provider='vkontakte')
        api = VkontakteApi()
        self.assertEqual(len(api.get_tokens()), 5)

        api.used_access_tokens = [token.access_token]
        for i in range(0, 100):
            self.assertNotEqual(api.get_token(), token.access_token)

        for i in range(0, 5):
            UserSocialAuth.objects.create(user=get_user_model().objects.create(username=i), uid=i,
                                          provider='vk-oauth2', extra_data='{"access_token": "qwewekrjhshe"}')
        self.assertEqual(len(api.get_tokens()), 10)

    @mock.patch('oauth_tokens.models.AccessToken.objects.fetch', side_effect=raise_error)
    def test_oauth_tokens_update_tokens(self, fetch):

        api = VkontakteApi()
        with self.assertRaises(CallsLimitError):
            api.update_tokens()

        self.assertTrue(fetch.called)
        self.assertEqual(fetch.call_count, 5)

        # TODO: when we update_tokens of new instance, counter is still max, that's wrong
        api = VkontakteApi()
        with self.assertRaises(CallsLimitError):
            api.update_tokens()
        # self.assertEqual(fetch.call_count, 10)



import time
from oauth_tokens.models import AccessToken, UserCredentials, AccessTokenGettingError, AccessTokenRefreshingError

from ..lock import distributedlock, LockNotAcquiredError
from ..utils import limit_errored_calls
from .base import TokensStorageAbstractBase


class OAuthTokensStorage(TokensStorageAbstractBase):

    name = 'oauth_tokens'

    def __init__(self, *args, **kwargs):
        super(OAuthTokensStorage, self).__init__(*args, **kwargs)
        self.tag = self.get_from_context('tag')
        self.only_this = bool(self.tag)

    def get_tokens(self):
        queryset = AccessToken.objects.filter(provider=self.provider).order_by('-granted_at')
        if self.tag:
            queryset = queryset.filter(user_credentials__in=UserCredentials.objects.filter(tags__name=self.tag))
        return queryset.values_list('access_token', flat=True)

    @limit_errored_calls(AccessTokenGettingError, 5)
    def update_tokens(self):
        lock_name = 'update_tokens_for_%s' % self.provider
        try:
            # the first call of method will update tokens, all others will just wait for releasing the lock
            with distributedlock(lock_name, blocking=False):
                # self.logger.info("Updating access tokens, method: %s, recursion count: %d" % (self.method,
                #                                                                               self.recursion_count))
                AccessToken.objects.fetch(provider=self.provider)
                return True
        except LockNotAcquiredError:
            # wait until lock will be released and return
            updated = False
            while not updated:
                # self.logger.info("Updating access tokens, waiting for another execution, method: %s, recursion "
                #                  "count: %d" % (self.method, self.recursion_count))
                try:
                    with distributedlock(lock_name, blocking=False):
                        updated = True
                except LockNotAcquiredError:
                    time.sleep(1)
            return True

    @limit_errored_calls(AccessTokenRefreshingError, 5)
    def refresh_tokens(self):
        # TODO: implement the same logic of distributedlock as in update_tokens method
        return AccessToken.objects.refresh(self.provider)



# distributedlock settings
import distributedlock
from django.core.cache import get_cache


distributedlock.DEFAULT_TIMEOUT = 60 * 5
distributedlock.DEFAULT_MEMCACHED_CLIENT = get_cache('default')

# import after distributedlock settings
from distributedlock import distributedlock, LockNotAcquiredError



# -*- coding: utf-8 -*-
import logging
import socket
import sys
import time
import random
import six
from six.moves.http_client import BadStatusLine, ResponseNotReady, IncompleteRead
from abc import ABCMeta, abstractmethod, abstractproperty
from ssl import SSLError

from requests.exceptions import ConnectionError
from django.conf import settings

from .exceptions import NoActiveTokens
from .utils import get_storages, override_api_context


__all__ = ['NoActiveTokens', 'ApiAbstractBase', 'Singleton', 'override_api_context']


class ApiAbstractBase(object):
    __metaclass__ = ABCMeta

    consistent_token = None
    error_class_repeat = (SSLError, ConnectionError, socket.error, BadStatusLine, ResponseNotReady, IncompleteRead)
    sleep_repeat_error_messages = []

    recursion_count = 0

    method = None
    used_access_tokens = None

    def __init__(self):
        self.tokens = []
        self.used_access_tokens = []
        self.consistent_token = None
        self.logger = self.get_logger()
        self.api = None

    def set_context(self):
        # define context of call on each calling, becouse instanse is singleton
        self.consistent_token = None

        context = getattr(settings, 'SOCIAL_API_CALL_CONTEXT', None)
        if context and self.provider in context and 'token' in context[self.provider]:
            self.consistent_token = context[self.provider]['token']

    def call(self, method, *args, **kwargs):
        self.method = method
        self.set_context()

        try:
            token = self.get_token()
        except NoActiveTokens as e:
            return self.handle_error_no_active_tokens(e, *args, **kwargs)

        self.api = self.get_api(token)

        try:
            response = self.get_api_response(*args, **kwargs)
        except self.error_class as e:
            response = self.handle_error_message(e, *args, **kwargs)
            if response is not None:
                return response
            response = self.handle_error_code(e, *args, **kwargs)
        except self.error_class_repeat as e:
            response = self.handle_error_repeat(e, *args, **kwargs)
        except Exception as e:
            return self.log_and_raise(e, *args, **kwargs)

        return response

    def handle_error_no_active_tokens(self, e, *args, **kwargs):
        if self.used_access_tokens:
            # wait 1 sec and repeat with empty used_access_tokens
            self.logger.warning("Waiting 1 sec, because all active tokens are used, method: %s, recursion count: %d" %
                                (self.method, self.recursion_count))
            self.used_access_tokens = []
            return self.sleep_repeat_call(*args, **kwargs)
        else:
            self.logger.warning("Suddenly updating tokens, because no active access tokens and used_access_tokens "
                                "empty, method: %s, recursion count: %d" % (self.method, self.recursion_count))
            self.update_tokens()
            return self.repeat_call(*args, **kwargs)

    def handle_error_message(self, e, *args, **kwargs):
        # check if error message contains any of defined messages
        for message in self.sleep_repeat_error_messages:
            if message in str(e):
                return self.sleep_repeat_call(*args, **kwargs)
        return

    def handle_error_code(self, e, *args, **kwargs):
        # try to find method for handling exception by it's code
        try:
            return getattr(self, 'handle_error_code_%s' % self.get_error_code(e))(e, *args, **kwargs)
        except AttributeError:
            return self.log_and_raise(e, *args, **kwargs)

    def log_and_raise(self, e, *args, **kwargs):
        self.logger.error("Error '%s'. Method %s, args: %s, kwargs: %s, recursion count: %d" % (
            e, self.method, args, kwargs, self.recursion_count))
        error_class = type(e)
        six.reraise(error_class, e, sys.exc_info()[2])

    def get_error_code(self, e):
        return e.code

    def handle_error_repeat(self, e, *args, **kwargs):
        self.logger.error("Exception: '%s' registered while executing method %s with params %s, recursion count: %d"
                          % (e, self.method, kwargs, self.recursion_count))
        return self.sleep_repeat_call(*args, **kwargs)

    def sleep_repeat_call(self, *args, **kwargs):
        time.sleep(kwargs.pop('seconds', 1))
        return self.repeat_call(*args, **kwargs)

    def repeat_call(self, *args, **kwargs):
        self.recursion_count += 1
        return self.call(self.method, *args, **kwargs)

    def update_tokens(self):
        self.consistent_token = None
        for storage in get_storages(self.provider):
            storage.update_tokens()

    def refresh_tokens(self):
        self.consistent_token = None
        for storage in get_storages(self.provider):
            storage.refresh_tokens()

    def get_tokens(self):
        tokens = []
        for storage in get_storages(self.provider):
            storage_tokens = storage.get_tokens()
            if storage.only_this:
                return storage_tokens
            tokens += storage_tokens
        return tokens

    def get_token(self):
        if self.consistent_token and self.consistent_token not in self.used_access_tokens:
            return self.consistent_token

        self.tokens = self.get_tokens()

        if not self.tokens:
            self.update_tokens()
            self.tokens = self.get_tokens()
            if not self.tokens:
                raise NoActiveTokens("There is no active tokens for provider %s after updating" % self.provider)

        tokens = self.tokens
        if self.used_access_tokens:
            tokens = list(set(tokens).difference(set(self.used_access_tokens)))
            if not tokens:
                raise NoActiveTokens("There is no active tokens for provider %s, used_tokens: %s"
                                     % (self.provider, self.used_access_tokens))

        return random.choice(tokens)

    def get_logger(self):
        return logging.getLogger('%s_api' % self.provider)

    @abstractproperty
    def provider(self):
        pass

    @abstractproperty
    def error_class(self):
        pass

    @abstractmethod
    def get_api(self, token):
        pass

    @abstractmethod
    def get_api_response(self):
        pass


class Singleton(ABCMeta):
    """
    Singleton metaclass for API classes
    from here http://stackoverflow.com/a/33201/87535
    """

    def __init__(cls, name, bases, dictionary):
        super(Singleton, cls).__init__(name, bases, dictionary)
        cls.instance = None

    def __call__(cls, *args, **kwargs):
        if cls.instance is None:
            cls.instance = super(Singleton, cls).__call__(*args, **kwargs)
        return cls.instance



# -*- coding: utf-8 -*-
from django.test import TestCase
from django.conf import settings


class SocialApiTestCase(TestCase):
    _settings = None
    token = None
    provider = ''

    def setUp(self):
        context = getattr(settings, 'SOCIAL_API_CALL_CONTEXT', {})
        self._settings = dict(context)
        context.update({self.provider: {'token': self.token}})

    def tearDown(self):
        setattr(settings, 'SOCIAL_API_CALL_CONTEXT', self._settings)



import logging
from abc import ABCMeta, abstractmethod, abstractproperty
from django.conf import settings


class TokensStorageAbstractBase(object):

    __metaclass__ = ABCMeta

    only_this = False

    @abstractproperty
    def name(self):
        pass

    @abstractmethod
    def get_tokens(self):
        pass

    @abstractmethod
    def update_tokens(self):
        pass

    @abstractmethod
    def refresh_tokens(self):
        pass

    def __init__(self, provider, *args, **kwargs):
        self.provider = provider
        self.logger = self.get_logger()

    def get_from_context(self, name):
        key = '_'.join([self.name, name])
        context = getattr(settings, 'SOCIAL_API_CALL_CONTEXT', None)
        if context and self.provider in context and key in context[self.provider]:
            return context[self.provider][key]

    def get_logger(self):
        return logging.getLogger('%s_api' % self.provider)




from social.apps.django_app.default.models import UserSocialAuth
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from .base import TokensStorageAbstractBase


PROVIDERS_MAP = getattr(settings, 'SOCIAL_API_SOCIAL_AUTH_PROVIDERS_MAP', {
    'vkontakte': 'vk-oauth2',
    'instagram': 'instagram',
    'facebook': 'facebook',
    'odnoklassniki': 'odnoklassniki-oauth2',
    'twitter': 'twitter',
})


class SocialAuthTokensStorage(TokensStorageAbstractBase):

    name = 'social_auth'

    def __init__(self, *args, **kwargs):
        super(SocialAuthTokensStorage, self).__init__(*args, **kwargs)
        self.user = self.get_from_context('user')
        self.only_this = bool(self.user)

    def get_provider(self):
        try:
            return PROVIDERS_MAP[self.provider]
        except KeyError:
            raise ImproperlyConfigured("Specify SOCIAL_API_SOCIAL_AUTH_PROVIDERS_MAP in settings "
                                       "with value for provider %s", self.provider)

    def get_tokens(self):
        queryset = UserSocialAuth.objects.filter(provider=self.get_provider())
        if self.user:
            queryset = queryset.filter(user=self.user)
        return [s.extra_data['access_token'] for s in queryset]

    def update_tokens(self):
        pass

    def refresh_tokens(self):
        pass



import time
from django.utils.module_loading import import_string
from django.core.exceptions import ImproperlyConfigured
from django.conf import settings
from django.test.utils import override_settings

from .storages.base import TokensStorageAbstractBase
from .exceptions import CallsLimitError


def get_storages_default():
    storages = getattr(settings, 'SOCIAL_API_TOKENS_STORAGES', None)
    if storages is None:
        storages_default = [
            ('social_api.storages.oauthtokens.OAuthTokensStorage', 'oauth_tokens'),
            ('social_api.storages.social_auth.SocialAuthTokensStorage', 'social.apps.django_app.default'),
        ]
        storages = [storage for storage, app in storages_default if app in settings.INSTALLED_APPS]
        if not storages:
            raise ImproperlyConfigured("No available token storages found for social_api application. Add to "
                                       "INSTALLES_APPS at least one storage: social_auth or oauth_tokens or provide "
                                       "custom setting SOCIAL_API_TOKENS_STORAGES")
    return storages


STORAGES = get_storages_default()


def get_storages(*args, **kwargs):
    for import_path in STORAGES:
        yield get_storage(import_path, *args, **kwargs)


def get_storage(import_path, *args, **kwargs):
    """
    Imports the staticfiles finder class described by import_path, where
    import_path is the full Python path to the class.
    """
    TokensStorage = import_string(import_path)
    if not issubclass(TokensStorage, TokensStorageAbstractBase):
        raise ImproperlyConfigured('TokensStorage "%s" is not a subclass of "%s"' % TokensStorageAbstractBase)
    return TokensStorage(*args, **kwargs)


def override_api_context(provider, **kwargs):
    context = getattr(settings, 'SOCIAL_API_CALL_CONTEXT', {provider: {}}).get(provider, {})
    context.update(kwargs)
    return override_settings(SOCIAL_API_CALL_CONTEXT={provider: context})


def limit_errored_calls(error, limit):

    def _inner_decorator(fn):
        fn.count = 1

        def _inner_function(*args, **kwargs):
            try:
                return fn(*args, **kwargs)
            except error:
                if fn.count < limit:
                    time.sleep(1)
                    fn.count += 1
                    return _inner_function(*args, **kwargs)
                else:
                    raise CallsLimitError("Limit of calls %s method %s achieved" % (limit, fn))

        return _inner_function

    return _inner_decorator
